/********************************************************
 * nup -- Print frame maker documents N up		*
 *	Where N is 2, or 4 logical pages per physical	*
 *	pages.						*
 *							*
 * Usage:						*
 *	nup -2 <files>					*
 * or							*
 *	nup -4 <files>					*
 *							*
 * Output is to standard out.				*
 *							*
 * Note: Error handling consists mainly of aborting	*
 ********************************************************/
#include <fstream.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#ifndef __GNUC__
typedef unsigned int bool;	// If your compiler doesn't have bool yet
static unsigned int true = 1;	// True value
static unsigned int false = 0;	// False value
#endif /* __GNUC__ */

static ifstream in_file;	// File to read

static const int LINE_SIZE = 1024;	// Longest possible line
static char line[LINE_SIZE];	// Line from the input

// 2 Up printing, first page
const char UP_2_PART_1[] = 
   "gsave\n"
   "0.70 0.70 scale\n"
   "0 350 translate\n";

// 2 Up printing, second page
const char UP_2_PART_2[] = 
   "gsave\n"
   "0.70 0.70 scale\n"
   "0 -190 translate\n";

// 4 Up printing, first page
const char UP_4_PART_1[] = 
   "gsave\n"
   "0.5 0.5 scale\n"
   "0 0 translate\n";

// 4 Up printing, second page
const char UP_4_PART_2[] =
   "gsave\n"
   "0.5 0.5 scale\n"
   "0 720 translate\n";

// 4 Up printing, third page
const char UP_4_PART_3[] = 
   "gsave\n"
   "0.5 0.5 scale\n"
   "600 0 translate\n";

// 4 Up printing, fourth page
const char UP_4_PART_4[] = 
   "gsave\n"
   "0.5 0.5 scale\n"
   "600 720 translate\n";

// Framemaker's showpage command
const char FM_SHOWPAGE[] = "	showpage";
const char FM_NO_SHOWPAGE[] = "%----removed	showpage";
const char END_SETUP[] = "%%EndSetup";
const char PAGE[] = "%%Page:";
const char END_PAGE[] = "%%EndPage";

// Begin page as we see it
const char FM_BEGIN[] =     "612 792 ? FMBEGINPAGE";
// Scanf/printf string for FMBEGIN
const char FM_SCAN_BEGIN[] = "612 792 %d FMBEGINPAGE";

/********************************************************
 * copy_header -- Copy the header			*
 *	In the process remove any "showpage"		*
 *	commands from the /FMENDPAGE script.		*
 ********************************************************/
static void copy_header(void)
{
    while (1) {
	if (in_file.eof()) return;
	if (in_file.bad()) return;

	line[0] = '\0';
	in_file.getline(line, sizeof(line));

	if (strcmp(line, FM_SHOWPAGE) == 0)
	    cout << FM_NO_SHOWPAGE << '\n';
	else
	    cout << line << '\n';

	if (strcmp(line, END_SETUP) == 0) {
	    return;
	}
    }
}
/********************************************************
 * begins -- Return true if one string begins another	*
 ********************************************************/
static bool begins(
   const char string[],		// String to look at
   const char q_string[]	// String that may begin it
)
{
   return (strncmp(string, q_string, strlen(q_string)) == 0);
}
/********************************************************
 * strqcmp -- like strcmp but allows ? in second string	*
 *							*
 * Returns						*
 *	0 == strings equal				*
 *	<0 == string1 < string2				*
 *	>0 == string1 > string2				*
 ********************************************************/
static int strqcmp(
   const char str1[],	// String to check
   const char str2[]	// Other string to check (may contain ?)
)
{
    register const char *ptr_1 = str1;	// First string
    register const char *ptr_2 = str2;	// Second string

    while (1) {
	if ((*ptr_1 == '\0') || (*ptr_2 == '\0'))
	    return ((*ptr_1) - (*ptr_2));
	
	if (*ptr_2 != '?') {
	    int diff = ((*ptr_1) - (*ptr_2));
	    if (diff != 0)
		return (diff);
	}
	++ptr_1;
	++ptr_2;
    }
}

/********************************************************
 * Copy page.						*
 *	Copy a single page.				*
 *							*
 * Returns:						*
 *	True -- page copied				*
 *	False -- Out of pages (line contains next	*
 *		line of input file)			*
 ********************************************************/
static bool copy_page(
   const char start_string[],	// Data to add at the beginning
   const bool un_rotate		// Unrotate the pages

)
{
    if (in_file.eof()) return (false);
    if (in_file.bad()) return (false);

    line[0] = '\0';
    in_file.getline(line, sizeof(line));

    if (! begins(line, PAGE)) return (false);
    cout << "%-----removed " << line << '\n';

    cout << start_string;	// Start the ball rolling


    while (1) {
	if (in_file.eof()) return (false);
	if (in_file.bad()) return (false);

	line[0] = '\0';
	in_file.getline(line, sizeof(line));

	if (un_rotate && (strqcmp(line, FM_BEGIN) == 0)) {
	    cout << "%---- removed " << line << '\n';
	    int flag;	// Flag from FM_BEGIN

	    sscanf(line, FM_SCAN_BEGIN, &flag);

	    if (!flag)
	       cout << "0 190 translate\n";

	    // The (1-flag) takes flag (which is 1 or 0)
	    // and toggles it to the other value
	    sprintf(line, FM_SCAN_BEGIN, 1 - flag);
	    cout << line << '\n';
	    continue;
	}

	if (begins(line, END_PAGE)) {
	    cout << "grestore\n";
	    cout << "%----removed " << line << '\n';
	    return (true);
	}
	cout << line << '\n';
    }
}

/********************************************************
 * copy_trailer -- Copy till the end of the file	*
 *	(First line in the buffer already)		*
 ********************************************************/
static void copy_trailer(void)
{
    while (1) {
	cout << line << '\n';

	if (in_file.eof()) return;
	if (in_file.bad()) return;

	line[0] = '\0';
	in_file.getline(line, sizeof(line));
    }
}

static int page_number;	// current page number for output page

/********************************************************
 * finish_page -- Put the showpage and other junk	*
 *	on the last page of the set.			*
 ********************************************************/
static void finish_page(void)
{
    cout << "showpage\n";
    cout << END_PAGE << '"' << page_number << '"' << 
		    ' ' << page_number << '\n';
    page_number++;
}

/********************************************************
 * do_2_up -- print pages 2 up.				*
 ********************************************************/
static void do_2_up(void)
{
    while (1) {
	cout << PAGE << '"' << page_number << '"' << 
			' ' << page_number << '\n';

	if (!copy_page(UP_2_PART_1, true)) break;
	if (!copy_page(UP_2_PART_2, true)) break;
	finish_page();
    }   
    finish_page();
}

/********************************************************
 * do_4_up -- print pages 4 up.				*
 ********************************************************/
static void do_4_up(void)
{
    while (1) {
	cout << PAGE << '"' << page_number << '"' << 
			' ' << page_number << '\n';

	if (!copy_page(UP_4_PART_1, false)) break;
	if (!copy_page(UP_4_PART_2, false)) break;
	if (!copy_page(UP_4_PART_3, false)) break;
	if (!copy_page(UP_4_PART_4, false)) break;
	finish_page();
    }   
    finish_page();
}

main(int argc, char *argv[])
{
    int n_up = 4;	// Default to 4 up printing

    while ((argc > 1) && (argv[1][0] == '-')) {
        switch (argv[1][1]) {
            case '4':
                n_up = 4;
                break;
	    case '2':
	        n_up = 2;
	        break;
	    default:
	        cerr << "Bad option " << argv[1] << '\n';
	        cerr << "Usage is   n_up [-2|-4] <files>\n";
		exit (8);
	}
	argc--;
	argv++;
    }

    while (argc > 1) {

	in_file.close();
	in_file.open(argv[1]);
	if (in_file.bad()) {
	    cerr << "Error: Could not open " << argv[1] << " for input\n";
	    exit (8);
	}

	copy_header();
	page_number = 1;	// Current page number

	switch (n_up) {
	    case 2:
		do_2_up();
		break;
	    case 4:
		do_4_up();
		break;
	    default:
	        cerr << "Internal error: n_up strange\n";
		exit (8);
	}
	copy_trailer();
	argc--;
	argv++;
    }

    return (0);
}

